/*
 *
 * Copyright (c) 2007 Steve Bennett <steveb@workware.net.au>
 *
 * A simple daemon which listens on a unix domain socket
 * and thus allows multiple clients to communicate with a single
 * topfield device.
*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <time.h>
#include <utime.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>

#include <tf_io.h>
#include <tf_util.h>
#include <mjd.h>
#include <tf_bytes.h>

#define LOG(FMT...) if (debug) fprintf(stderr, FMT)

static int devindex = 0;
static int trace_level = 0;
static int debug = 0;
static int topfield_is_open = 0;
static tf_handle tf;

static int sig_int = 0;
static int sig_hup = 0;

static int parse_args(int argc, char *argv[]);

static void sig_handle(int sig)
{
	if (sig == SIGHUP) {
		sig_hup = 1;
	}
	else if (sig == SIGPIPE) {
		/* ignore it */
	}
	else {
		sig_int = 1;
	}
	signal(sig, sig_handle);
}

#define MAX_CLIENTS 5

static int add_fd(fd_set *set, int fd, int maxfd)
{
	if (fd >= 0) {
		FD_SET(fd, set);
		if (fd > maxfd) {
			maxfd = fd;
		}
	}
	return maxfd;
}

static void
reply(int fd, int n, const char *fmt, ...)
{
	struct iovec vec[3];

	char code[5];
	char buf[512];

	va_list ap;
	va_start (ap, fmt);

	snprintf(code, sizeof(code), "%03d ", n);

	vsnprintf(buf, sizeof(buf), fmt, ap);

	LOG( "--> %03d %s\n", n, buf);

	vec[0].iov_base = code;
	vec[0].iov_len = strlen(code);
	vec[1].iov_base = buf;
	vec[1].iov_len = strlen(buf);
	vec[2].iov_base = "\n";
	vec[2].iov_len = 1;

	writev(fd, vec, 3);
}

static void
fdprintf(int fd, const char *fmt, ...)
{
	char buf[1024];

	va_list ap;
	va_start (ap, fmt);

	vsnprintf(buf, sizeof(buf), fmt, ap);

	write(fd, buf, strlen(buf));
}

static void
error_reply(int fd, int rc)
{
	reply(fd, 500, "Topfield error: %s", tf_strerror(rc));
}

static int open_topfield(int intr_transfer)
{
	if (!topfield_is_open) {
		int ret = topfield_open(&tf, devindex, TF_LOCK_HUP);

		LOG("topfield_open() returned %d\n", ret);

		if (ret > 0) {
			/* Failed to lock either because something bad happened or
			 * the existing locker refused to give up the lock
			 */
			//reply(550, "Topfield is already in use");
			return 0;
		}
		if (ret < 0) {
			//reply(550, "Failed to connect to Topfield PVR");
			return 0;
		}

		/* REVISIT: timeout */
		tf.trace_level = trace_level;
		/* REVISIT: Option to reenable this check */
		tf.nocrc = 1;

		if (tf_init(&tf) != 0) {
			topfield_close(&tf);
			//reply(550, "Failed to establish communication with Topfield");
			return 0;
		}

		topfield_is_open = 1;
	}
	return topfield_is_open;
}

static void close_topfield(void)
{
	LOG("close_topfield() topfield_is_open=%d\n", topfield_is_open);

	if (topfield_is_open) {
		/* We might have been in the middle of a get or put, so cancel anything if necessary */
		tf_cmd_cancel(&tf);

		topfield_close(&tf);
		topfield_is_open = 0;
	}
}

static char *unescape(char *pt, char **next)
{
	char *dest;
	char *result;

	//fprintf(stderr, "unescape(%s) -> ", pt);

	while (pt && *pt == ' ') {
		pt++;
	}
	if (!pt || !*pt) {
		//fprintf(stderr, "(null)\n");
		return 0;
	}
	result = dest = pt;
	while (*pt) {
		if (pt[0] == '\\' && pt[1]) {
			*dest++ = pt[1];
			pt += 2;
		}
		else {
			if (next && *pt == ' ') {
				/* Stop at the first unescaped space and store the pointer
				 * to the start of the next arg in *next */
				*next = pt + 1;
				break;
			}
			*dest++ = *pt++;
		}
	}
	*dest = 0;

	//fprintf(stderr, "%s\n", result);

	return result;
}

typedef int cmd_function_t(tf_handle *tf, int fd, char *args);

typedef struct {
	const char *name;
	int noconn;
	cmd_function_t *handler;
} cmd_t;

static int cmd_hello(tf_handle *tf, int fd, char *args)
{
	reply(fd, 200, "Hello");
	return 0;
}

static int cmd_bye(tf_handle *tf, int fd, char *args)
{
	reply(fd, 201, "Goodbye");
	close(fd);
	return 1;
}

static int cmd_size(tf_handle *tf, int fd, char *args)
{
	tf_size_result result;
	int rc;

	if ((rc = tf_cmd_size(tf, &result)) == 0) {
		reply(fd, 200, "%u %u", result.totalk, result.freek);
	}
	else {
		error_reply(fd, rc);
	}
	return 0;
}

static int cmd_mkdir(tf_handle *tf, int fd, char *args)
{
	int rc;
	args = unescape(args, 0);

	if (!args) {
		reply(fd, 500, "No argument to mkdir");
		return 0;
	}

	rc = tf_cmd_mkdir(tf, args);
	if ((rc = tf_cmd_mkdir(tf, args)) == 0) {
		reply(fd, 200, "Directory '%s' created", args);
	}
	else {
		error_reply(fd, rc);
	}
	return 0;
}

static int cmd_info(tf_handle *tf, int fd, char *args)
{
	tf_dirent dirent;
	tf_buffer buf;
	int rc;

	args = unescape(args, 0);

	if (!args) {
		reply(fd, 500, "No argument to info");
		return 0;
	}

	/* Fetch the first 4K of the file */
	if ((rc = tf_cmd_get(tf, args, 0, &dirent)) != 0) {
		error_reply(fd, rc);
		return 0;
	}

	rc = tf_cmd_get_next(tf, &buf);

	/* Now cancel the GET operation */
	tf_cmd_get_cancel(tf);

	if (rc != 0) {
		error_reply(fd, rc);
		return 0;
	}

	if (memcmp(buf.data, "TFrc", 4) != 0) {
		reply(fd, 500, "Not a REC file: %s", args);
		return 0;
	}
	else {
		/* Now extract some info from the buffer */
		struct tm *tm;
		char service_name[25];
		char event_name[256];
		char event_desc[256];
		struct tf_datetime event_start;
		time_t event_start_t;
		char event_start_str[50];
		char *event_extinfo;

		/* REVISIT: Different models have different offsets.
		 * How can I tell which model?
		 */

#if 0
		__u8 service_num = buf.data[10];
		__u8 service_type = buf.data[13];
		__u16 event_duration = get_u16(buf.data + 70);
#endif
		__u16 duration = get_u16(buf.data + 8);
		__u8 event_name_len = buf.data[85];
		__u16 event_extinfo_len = get_u16(buf.data + 362);

		event_start.mjd = get_u16(buf.data + 76);
		event_start.hour = buf.data[78];
		event_start.minute = buf.data[79];

		event_start_t = tfdt_to_time(&event_start);

		/* Ugly way to adjust for TZ */
		tm = localtime(&event_start_t);
		event_start_t += tm->tm_gmtoff;

		memcpy(service_name, buf.data + 28, 24);
		service_name[24] = 0;

		memcpy(event_name, buf.data + 87, event_name_len);
		event_name[event_name_len] = 0;

		memcpy(event_desc, buf.data + 87 + event_name_len, 255);
		event_desc[255] = 0;

		event_extinfo = malloc(event_extinfo_len + 1);
		memcpy(event_extinfo, buf.data + 368, event_extinfo_len);
		event_extinfo[event_extinfo_len] = 0;

		strftime(event_start_str, sizeof(event_start_str), "%H:%M %a, %d %b %Y", localtime(&event_start_t));

		/* Now output the results */
		fdprintf(fd, "\tduration\t%u\t%d:%02d\n", duration, duration / 60, duration % 60);
		fdprintf(fd, "\tchannel\t%s\n", service_name);
		fdprintf(fd, "\tstart\t%u\t%s\n", event_start_t, event_start_str);
		fdprintf(fd, "\tname\t%s\n", event_name);
		fdprintf(fd, "\tdesc\t%s\n", event_desc);
		fdprintf(fd, "\textinfo\t%s\n", event_extinfo);

		//printf("header='%c%c%c%c'\n", buf.data[0], buf.data[1], buf.data[2], buf.data[3]);

		free(event_extinfo);

		reply(fd, 200, "Info complete for %s", args);
	}
	return 0;
}

static int cmd_delete(tf_handle *tf, int fd, char *args)
{
	int rc;
	args = unescape(args, 0);

	if (!args) {
		reply(fd, 500, "No argument to delete");
		return 0;
	}

	if ((rc = tf_cmd_delete(tf, args)) == 0) {
		reply(fd, 200, "Deleted '%s'", args);
	}
	else {
		error_reply(fd, rc);
	}
	return 0;
}

static int cmd_rename(tf_handle *tf, int fd, char *args)
{
	int rc;
	char *arg1 = unescape(args, &args);
	char *arg2 = unescape(args, 0);

	if (!arg1 || !arg2) {
		reply(fd, 500, "rename requires two arguments");
		return 0;
	}

	if ((rc = tf_cmd_rename(tf, arg1, arg2)) == 0) {
		reply(fd, 200, "Renamed '%s' to '%s'", arg1, arg2);
	}
	else {
		error_reply(fd, rc);
	}
	return 0;
}

static int cmd_dir(tf_handle *tf, int fd, char *args)
{
	int rc;
	tf_dir_entries entries;

	args = unescape(args, 0);

	if (!args) {
		args = "/";
	}

	if ((rc = tf_cmd_dir_first(tf, args, &entries)) == 0) {
		do {
			int i;

			for (i = 0; i < entries.count; i++) {
				/* Response is tab separated with an initial tab */
				fdprintf(fd, "\t%ld\t%s\t%lld\t%s\n",
					(long)entries.entry[i].stamp,
					entries.entry[i].type == 'f' ? "file" : "dir",
					entries.entry[i].size,
					entries.entry[i].name);
			}
		} while ((rc = tf_cmd_dir_next(tf, &entries)) == 0);

		if (rc == TF_ERR_DONE) {
			reply(fd, 200, "Directory listing complete");
		}
		else {
			error_reply(fd, rc);
		}
	}
	return 0;
}

static const cmd_t cmd_table[] = {
	{
		.name = "hello",
		.noconn = 1,
		.handler = cmd_hello,
	},
	{
		.name = "bye",
		.noconn = 1,
		.handler = cmd_bye,
	},
	{
		.name = "size",
		.handler = cmd_size,
	},
	{
		.name = "mkdir",
		.handler = cmd_mkdir,
	},
	{
		.name = "info",
		.handler = cmd_info,
	},
	{
		.name = "delete",
		.handler = cmd_delete,
	},
	{
		.name = "rename",
		.handler = cmd_rename,
	},
	{
		.name = "dir",
		.handler = cmd_dir,
	},
	{
		.name = 0,
	}
};

int main(int argc, char *argv[])
{
	struct sockaddr_un addr;
	int clients[MAX_CLIENTS];
	int s;
	int i;

	if (parse_args(argc, argv) != 0) {
		return 1;
	}

	/* Create the unix domain socket */
	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s < 0) {
		perror("socket");
		return 1;
	}

	/* And bind to /tmp/toppyd */
	/* REVISIT: Really should be toppyd.<devindex> */
	remove("/tmp/toppyd");
	strcpy(addr.sun_path, "/tmp/toppyd");
	addr.sun_family = AF_UNIX;
	if (bind(s, (struct sockaddr *)&addr, strlen(addr.sun_path) + 1 + sizeof(addr.sun_family)) != 0) {
		perror("bind");
		return 1;
	}

	listen(s, 5);

	signal(SIGINT, sig_handle);
	signal(SIGTERM, sig_handle);
	signal(SIGHUP, sig_handle);
	signal(SIGPIPE, sig_handle);

	for (i = 0; i < MAX_CLIENTS; i++) {
		clients[i] = -1;
	}

	while (!sig_int) {
		fd_set readers;
		int maxfd;
		int rc;

		/* If we got SIG_HUP, give up the connection */
		if (sig_hup) {
			close_topfield();
			sig_hup = 0;
		}

		FD_ZERO(&readers);

		/* Wait for a connection or a client request */
		maxfd = add_fd(&readers, s, 0);
		for (i = 0; i < MAX_CLIENTS; i++) {
			maxfd = add_fd(&readers, clients[i], maxfd);
		}

		rc = select(maxfd + 1, &readers, 0, 0, 0);
		if (rc > 0) {
			int index;
			for (index = 0; index < MAX_CLIENTS; index++) {
				int fd = clients[index];

				if (fd >= 0 && FD_ISSET(fd, &readers)) {
					char buf[1024];
					int len;
					char *args;
					int found = 0;
					
					len = read(fd, buf, sizeof(buf));
					if (len <= 0) {
						LOG("client[%d] has gone away\n", index);
						close(fd);
						clients[index] = -1;
						continue;
					}

					if (buf[len - 1] == '\n') {
						len--;
					}
					buf[len] = 0;
					LOG("Got message from client[%d]: '%s'\n", index, buf);

					args = strchr(buf, ' ');
					if (args) {
						*args++ = 0;
					}

					for (i = 0; cmd_table[i].name; i++) {
						if (strcasecmp(buf, cmd_table[i].name) == 0) {
							int rc;

							found = 1;
							
							LOG("Found matching request: cmd=%s args='%s'\n", buf, args);

							if (!cmd_table[i].noconn) {
								/* Need a connection for this command */
								if (open_topfield(0) == 0) {
									LOG( "Failed to open connection to topfield\n");
									reply(fd, 501, "Failed connection - Could not connect to topfield %d", devindex);
									break;
								}
							}

							rc = cmd_table[i].handler(&tf, fd, args);
							LOG("Command returned %d\n", rc);
							if (rc == 1) {
								/* This means to close the connection */
								clients[index] = -1;
							}
							break;
						}
					}

					if (!found) {
						reply(fd, 500, "Unknown command: '%s'", buf);
					}
				}
			}
			if (FD_ISSET(s, &readers)) {
				struct sockaddr_un from;
				socklen_t fromlen = sizeof(from);
				int clientfd = accept(s, (struct sockaddr *)&from, &fromlen);
				if (clientfd > 0) {
					int slot = -1;
					for (i = 0; i < MAX_CLIENTS; i++) {
						if (clients[i] == -1) {
							slot = i;
							break;
						}
					}
					if (slot >= 0) {
						clients[slot] = clientfd;
						LOG( "New client[%d]\n", slot);
					}
					else {
						LOG( "Too many clients. Connection refused\n");
						reply(clientfd, 500, "Connection refused -- Too many clients");
						close(clientfd);
					}
				}
			}
		}
		else if (errno != EINTR) {
			perror("select");
		}
	}

	close_topfield();

	return 0;
}

void usage(void)
{
	fprintf(stderr, 
		"Usage: toppyd [-pP] [-D] [-d <devicenum>]\n"
		" -p			 - partial trace\n"
		" -P			 - full trace\n"
		" -D			 - debug\n"
		" -d <index>	 - device index, if you have more than one toppy\n"
	);
}

static int parse_args(int argc, char *argv[])
{
	int c;

	while((c = getopt(argc, argv, "pPd:D")) != -1) {
		switch (c)
		{
			case 'p':
				trace_level = 1;
				break;

			case 'P':
				trace_level = 2;
				break;

			case 'd':
				devindex = atoi(optarg);
				break;

			case 'D':
				debug++;
				break;

			default:
				usage();
				return -1;
		}
	}

	if (optind != argc) {
		usage();
		return -1;
	}

	return 0;
}
